@@ -1,10 +1,11 @@
 import asyncio
 from asyncio     import Future, PriorityQueue
-from typing      import (Awaitable, Deque, Dict, List, Optional, Set, Tuple,
-    Union)
+from typing      import (AsyncIterable, Awaitable, Deque, Dict, Iterable, List,
+    Optional, Set, Tuple, Union)
 from collections import deque
 from time        import monotonic
 
+import anyio
 from asyncio_throttle   import Throttler
 from async_timeout      import timeout
 from ircstates          import Emit, Channel, ChannelUser
@@ -15,7 +16,6 @@ from irctokens          import build, Line, tokenise
 from .ircv3     import (CAPContext, sts_transmute, CAP_ECHO, CAP_SASL,
     CAP_LABEL, LABEL_TAG_MAP, resume_transmute)
 from .sasl      import SASLContext, SASLResult
-from .join_info import WHOContext
 from .matching  import (ResponseOr, Responses, Response, ANY, SELF, MASK_SELF,
     Folded)
 from .asyncs    import MaybeAwait, WaitFor
@@ -25,9 +25,10 @@ from .interface import (IBot, ICapability, IServer, SentLine, SendPriority,
     IMatchResponse)
 from .interface import ITCPTransport, ITCPReader, ITCPWriter
 
-THROTTLE_RATE = 4 # lines
-THROTTLE_TIME = 2 # seconds
+THROTTLE_RATE = 4  # lines
+THROTTLE_TIME = 2  # seconds
 PING_TIMEOUT  = 60 # seconds
+WAIT_TIMEOUT  = 20 # seconds
 
 JOIN_ERR_FIRST = [
     ERR_NOSUCHCHANNEL,
@@ -64,8 +65,10 @@ class Server(IServer):
 
         self._read_queue:  Deque[Tuple[Line, Optional[Emit]]] = deque()
 
-        self._wait_fors: List[Tuple[WaitFor, Optional[Awaitable]]] = []
-        self._wait_for_fut: Optional[Future[WaitFor]] = None
+        self._wait_fors: List[WaitFor] = []
+        self._wait_for_fut: Dict[asyncio.Task, Future[bool]] = {}
+
+        self._pending_who: Deque[str] = deque()
 
     def hostmask(self) -> str:
         hostmask = self.nickname
@@ -179,11 +182,26 @@ class Server(IServer):
             elif emit.command == "JOIN":
                 if emit.self and not emit.channel is None:
                     await self.send(build("MODE", [emit.channel.name]))
-                    await WHOContext(self).ensure(emit.channel.name)
+
+                    self._pending_who.append(emit.channel.name)
+                    if len(self._pending_who) == 1:
+                        await self._serial_who()
 
         await self.line_read(line)
 
-    async def _next_lines(self) -> List[Tuple[Line, Optional[Emit]]]:
+    async def _serial_who(self):
+        while self._pending_who:
+            next = self._pending_who.popleft()
+            if self.isupport.whox:
+                await self.send(self.prepare_whox(next))
+            else:
+                await self.send(build("WHO", [next]))
+
+            end = Response(RPL_ENDOFWHO, [ANY, Folded(next)])
+            line = await self.wait_for(end)
+
+
+    async def _next_lines(self) -> AsyncIterable[Tuple[Line, Optional[Emit]]]:
         ping_sent = False
         while True:
             try:
@@ -195,10 +213,10 @@ class Server(IServer):
                 else:
                     ping_sent = True
                     await self.send(build("PING", ["hello"]))
-                continue
+                    continue
 
-            self.last_read  = monotonic()
-            ping_sent       = False
+            self.last_read = monotonic()
+            ping_sent      = False
 
             try:
                 lines = self.recv(data)
@@ -206,45 +224,45 @@ class Server(IServer):
                 self.disconnected = True
                 raise
 
-            return lines
+            for both in lines:
+                yield both
 
     async def _line_or_wait(self,
-            line_aw: Awaitable
-            ) -> Optional[Tuple[WaitFor, Awaitable]]:
-        wait_for_fut: Future[WaitFor] = Future()
-        self._wait_for_fut = wait_for_fut
+            line_aw: asyncio.Task
+            ) -> Optional[Awaitable]:
+        wait_for_fut: Future[bool] = Future()
+        self._wait_for_fut[line_aw] = wait_for_fut
 
         done, pend = await asyncio.wait([line_aw, wait_for_fut],
             return_when=asyncio.FIRST_COMPLETED)
-        self._wait_for_fut = None
+        del self._wait_for_fut[line_aw]
 
         if wait_for_fut.done():
             new_line_aw = list(pend)[0]
-            return (await wait_for_fut), new_line_aw
+            return new_line_aw
         else:
             return None
 
-    async def _read_lines(self) -> List[Tuple[Line, Optional[Emit]]]:
-        lines = await self._next_lines()
-        for line, emit in lines:
-            self.line_preread(line)
+    async def _read_lines(self) -> AsyncIterable[Tuple[Line, Optional[Emit]]]:
+        async with anyio.create_task_group() as tg:
+            async for both in self._next_lines():
+                line, emit = both
+                self.line_preread(line)
 
-            for i, (wait_for, aw) in enumerate(self._wait_fors):
-                if wait_for.match(self, line):
-                    wait_for.resolve(line)
+                for i, wait_for in enumerate(self._wait_fors):
+                    if wait_for.match(self, line):
+                        wait_for.resolve(line)
+                        self._wait_fors.pop(i)
+                        break
 
-                    if aw is not None:
-                        new_wait_for = await self._line_or_wait(aw)
-                        if new_wait_for is not None:
-                            self._wait_fors.append(new_wait_for)
-                    self._wait_fors.pop(i)
-                    break
+                line_aw = asyncio.create_task(self._on_read(line, emit))
+                new_wait = await self._line_or_wait(line_aw)
+                if new_wait is not None:
+                    async def _aw():
+                        await new_wait
+                    await tg.spawn(_aw)
 
-            line_aw = self._on_read(line, emit)
-            new_wait_for = await self._line_or_wait(line_aw)
-            if new_wait_for is not None:
-                self._wait_fors.append(new_wait_for)
-        return lines
+                yield both
 
     async def wait_for(self,
             response: Union[IMatchResponse, Set[IMatchResponse]],
@@ -258,20 +276,20 @@ class Server(IServer):
             response_obj = response
 
         our_wait_for = WaitFor(response_obj)
+        self._wait_fors.append(our_wait_for)
 
-        wait_for_fut = self._wait_for_fut
-        if wait_for_fut is not None:
-            self._wait_for_fut = None
-            wait_for_fut.set_result(our_wait_for)
-        else:
-            self._wait_fors.append((our_wait_for, None))
+        cur_task = asyncio.current_task()
+        if cur_task is not None and cur_task in self._wait_for_fut:
+            wait_for_fut = self._wait_for_fut[cur_task]
+            wait_for_fut.set_result(True)
 
         if sent_aw is not None:
             sent_line = await sent_aw
             label = str(sent_line.id)
             our_wait_for.with_label(label)
 
-        return (await our_wait_for)
+        async with timeout(WAIT_TIMEOUT):
+            return (await our_wait_for)
 
     async def _on_send_line(self, line: Line):
         if (line.command == "PRIVMSG" and
